;;; What needs to be done here is
;;; 1. to extend the element type to include a deferred expression
;;; 2. use that extra type to do incremental reading
;;; 3. add a proper class_print procedure

compile_mode :pop11 +strict;

section;

;;; -- UTILITIES ----------------------------------------------

;;; Usual disgusting hack to protect against defect in "uses".
vars xml = true;


;;; These routines would really be provided as autoloadable routines.
;;; They are not a required part of the XML package.

lconstant white_chars = '\s\t\n\r';

define is_white_string( x );
    returnunless( x.isstring )( false );
    lvars ch;
    for ch in_string x do
        returnunless( locchar( ch, 1, white_chars ) )( false )
    endfor;
    true
enddefine;

;;; We really need a Perl-like capability for this.  We want to
;;; write a one-liner like .... /\w*([^\w]*)\w*/$1/
;;; This is a mess really.  But it does the job.
;;;
define trim_white_string( x, start, end );
    check_string( x );
    lvars lo = 1, hi = datalength( x ), len = hi;
    if start then
        until lo fi_> len do
            quitunless( locchar( fast_subscrs( lo, x ), 1, white_chars ) );
            lo fi_+ 1 -> lo
        enduntil;
    endif;
    if end then
        until lo fi_> hi do
            quitunless( locchar( fast_subscrs( hi, x ), 1, white_chars ) );
            hi fi_- 1 -> hi
        enduntil;
    endif;
    if lo == 1 and hi == len then
        x
    elseif lo fi_> hi then
        nullstring
    else
        substring( lo, hi fi_- lo fi_+ 1, x )
    endif
enddefine;



;;; -- XML/HTML Datatype --------------------------------------

;;; This is used purely internally.  The name/value list is
;;; actually represented as a sorted list of Attributes.
;;;
defclass lconstant Attribute {
    attributeName,
    attributeValue
};

;;; We position the children at the end so that the idiom
;;;     e.destXmlElement.p
;;; works nicely.  In other words, after blowing all the bits
;;; onto the stack the vector-of-children is left nicely exposed
;;; on the top.
;;;
defclass lconstant XmlElement {
    xmlElementType,
    xmlElementAttributes,
    xmlElementChildren
};

define lconstant resort_attr_list( L );
    nc_listsort(
        L,
        procedure( x, y );
            alphabefore( attributeName( x ), attributeName( y ) )
        endprocedure
    )
enddefine;

define lconstant to_attr_list( L );
    [%
        until null( L ) do
            consAttribute( dest( fast_destpair( L ) ) -> L )
        enduntil
    %].resort_attr_list
enddefine;

define new_element( t );
    lvars n, attrs;
    if t.isword then
        [] -> attrs;
        consvector() -> n;
    elseif t.islist then
        t -> attrs;
        () -> t;
        consvector() -> n;
    elseif t.isinteger then
        consvector( t ) -> n;
        () -> t;
        if t.isword then
            [] -> attrs
        elseif t.islist then
            t -> attrs;
            () -> t
        else
            mishap( 'Unexpected arguments for new_element', [ ^t ^n ] )
        endif
    else
        mishap( 'Invalid argument', [ ^t ] )
    endif;
    consXmlElement( t, attrs.to_attr_list, n )
enddefine;


;;; define new_element( n, type );
;;;     lvars attrs = [];
;;;     if type.islist then
;;;         ( n, type ) -> ( n, type, attrs )
;;;     endif;
;;;     consvector( n ) -> n;
;;;     consXmlElement(
;;;         type,
;;;         attrs.to_attr_list,
;;;         n
;;;     )
;;; enddefine;

define element_type =
    xmlElementType
enddefine;

define element_attributes( e );
    applist( xmlElementAttributes( e ), destAttribute )
enddefine;

define element_attributes_list( e );
    maplist( xmlElementAttributes( e ), destAttribute )
enddefine;

define element_value( name, e );
    lvars attrs = xmlElementAttributes( e );
    lvars i;
    for i in attrs do
        returnif( attributeName( i ) == name )( attributeValue( i ) )
    endfor;
    false
enddefine;

;;; This bizarre implementation is chosen because node COPYING is
;;; more common than updating attributes.  Therefore we copy on
;;; write, so to speak.  An improvement would be to add a dirty mark
;;; to nodes.  Not worth it at the moment.
;;;
define updaterof element_value( value, name, e );
    lvars attrs = xmlElementAttributes( e );
    [%
        if value then consAttribute( name, value ) endif;
        lvars i;
        for i in attrs do
            unless attributeName( i ) == name do
                i
            endunless
        endfor;
    %].resort_attr_list -> xmlElementAttributes( e )
enddefine;

;;; This loop provides the syntax
;;;     for name, value in_attributes element do ... endfor
;;; The fast variant is not used.
;;;
define :for_extension in_attributes( varlist, isfast );
    dlocal pop_new_lvar_list;

    ;;; This is a shortcoming that will be addressed at a later date.
    unless varlist.length == 2 do
        mishap( varlist, 1, 'Exactly 2 variables required for in_attributes' )
    endunless;

    lvars ( namev, valuev ) = varlist.dl;
    lvars attrlistv = sysNEW_LVAR();

    pop11_comp_expr_to( "do" ) -> _;
    sysCALLQ( xmlElementAttributes );
    sysPOP( attrlistv );

    lvars start_label = sysNEW_LABEL().dup.pop11_loop_start;
    lvars done_label = sysNEW_LABEL().dup.pop11_loop_end;

    sysLABEL( start_label );

    sysPUSH( attrlistv );
    sysPUSHQ( [] );
    sysCALL( "==" );
    sysIFSO( done_label );
    sysPUSH( attrlistv );
    sysCALL( "fast_destpair" );
    sysPOP( attrlistv );
    sysCALLQ( destAttribute );
    sysPOP( valuev );
    sysPOP( namev );

    pop11_comp_stmnt_seq_to( "endfor" ) -> _;

    sysGOTO( start_label );
    sysLABEL( done_label );
enddefine;


;;; This loop provides the syntax
;;;     for child in_children element do ... endfor
;;; The fast variant is not used.
;;;
define :for_extension in_children( varlist, isfast );
    dlocal pop_new_lvar_list;

    ;;; This is a shortcoming that will be addressed at a later date.
    unless varlist.length == 1 do
        mishap( varlist, 1, 'Only one variable allowed in in_children' )
    endunless;
    lvars lv = varlist( 1 );

    lvars v = sysNEW_LVAR();
    lvars i = sysNEW_LVAR();
    lvars n = sysNEW_LVAR();
    pop11_comp_expr_to( "do" ) -> _;
    lvars start_label = sysNEW_LABEL().dup.pop11_loop_start;
    lvars done_label = sysNEW_LABEL().dup.pop11_loop_end;

    sysCALLQ( xmlElementChildren );
    sysPOP( v );
    sysPUSH( v );
    sysCALL( "datalength" );
    sysPOP( n );
    sysPUSHQ( 0 );
    sysPOP( i );

    sysLABEL( start_label );

    sysPUSH( i );
    sysPUSHQ( 1 );
    sysCALL( "fi_+" );
    sysPUSHS( i );
    sysPOP( i );
    sysPUSH( n );
    sysCALL( "fi_<=" );
    sysIFNOT( done_label );

    sysPUSH( i );
    sysPUSH( v );
    ;;; sysCALL( "fast_subscrv" );
    sysFIELD( false, conspair( "full", _ ), false, false );
    sysPOP( lv );

    pop11_comp_stmnt_seq_to( "endfor" ) -> _;

    sysGOTO( start_label );
    sysLABEL( done_label );
enddefine;

define app_element_children( x, p );
    appdata( x.xmlElementChildren, p )
enddefine;

define map_element_children( x, p );
    consXmlElement(
        x.destXmlElement -> x,
        {% appdata( x, p ) %}
    )
enddefine;

define element_length( x );
    x.xmlElementChildren.datalength
enddefine;

define element_last( x );
    x.xmlElementChildren.last
enddefine;

define element_first( x );
    x.xmlElementChildren.first
enddefine;

define subscr_element( n, x );
    subscrv( n, xmlElementChildren( x ) )
enddefine;

define updaterof subscr_element( v, n, x );
    v -> subscrv( n, xmlElementChildren( x ) )
enddefine;

define element_explode( x );
    x.xmlElementChildren.explode
enddefine;

define updaterof element_explode( x );
    -> x.xmlElementChildren.explode
enddefine;

define element_children( x );
    x.xmlElementChildren.explode
enddefine;

define element_children_list( x );
    x.xmlElementChildren.destvector.conslist
enddefine;

define element_children_dest( x );
    x.xmlElementChildren.destvector
enddefine;

define is_element =
    isXmlElement
enddefine;

;;; Note that you have to copy the vector of children.  This is
;;; highly undesirable.  What we what is a modified updater of
;;; subscr_element that marks elements as dirty.  If the element
;;; is clean then copying is simply copy(e).
;;;
define copy_element( e );
    consXmlElement( e.destXmlElement.copy )
enddefine;


;;; -- HTML Predicates ----------------------------------------

;;;
;;; is_stand_alone
;;; is_optionally_closed
;;; may_contain
;;;

lconstant standalone =
    [
        BASE LINK META HR INPUT COL FRAME PARAM IMG IFRAME
        BR SPACER WBR BASEFONT BGSOUND AREA
    ];

define is_stand_alone( type );
    lmember( type, standalone ) and true
enddefine;


lconstant physical_markup =
    [ B BDO BIG BLINK FONT I MARQUEE NOBR S SMALL SPAN STRIKE SUB SUP TT U ];

lconstant logical_markup =
    [ ABBR CITE CODE DFN EM KBD Q SAMP STRONG VAR ];

lconstant phrase_markup = physical_markup <> logical_markup;

lconstant body_stuff =
    [
        A APPLET BASEFONT
        BR BUTTON IFRAME IMG INPUT
        LABEL MAP OBJECT SCRIPT SELECT
        TEXTAREA
    ];

lconstant para_stuff = phrase_markup <> body_stuff;

lconstant body_blocks =
    [
        ADDRESS BLOCKQUOTE CENTER DIR DIV DL
        FIELDSET FORM H1 H2 H3 H4 H5 H6 HR ISINDEX MENU
        NOFRAMES NOSCRIPT OL P PRE TABLE UL
    ];

lconstant body_context = [ ^^body_blocks ^^para_stuff ];

define lconstant optional =
    newproperty(
        [
            [ P ^para_stuff ]
            [ OPTION [] ]
            [ DT ^para_stuff ]
            [ DD ^para_stuff ]
            [ LI ^body_context ]
            [ TBODY [TR] ]
            [ TFOOT [TR] ]
            [ THEAD [TR] ]
            [ TR [TH TD] ]
            [ TD ^body_context ]
            [ TH ^body_context ]
        ].dup.length,
        false,
        "perm"
    )
enddefine;

define is_optionally_closed( type );
    optional( type ) and true
enddefine;


define lconstant is_html =
    newproperty(
        maplist(
            [
                A ABBR ADDRESS APPLET AREA B BASE BASEFONT BDO BGSOUND BIG BLINK BLOCKQUOTE BODY
                BR BUTTON CAPTION CENTER CITE CODE COL COLGROUP DD DFN DIR DIV DL DT EM EMBED
                FIELDSET FONT FORM FRAME FRAMESET HEAD H1 H2 H3 H4 H5 H6 HR HTML I IFRAME ILAYER IMG INPUT
                ISINDEX KBD KEYGEN LABEL LAYER LEGEND LI LINK MAP MARQUEE MENU META NOBR
                NOEMBED NOFRAMES NOLAYER NOSCRIPT OBJECT OL OPTGROUP OPTION P PARAM PRE Q S
                SAMP SCRIPT SELECT SERVER SMALL SPACER SPAN STRIKE STRONG STYLE SUB SUP TABLE
                TBODY TD TEXTAREA TFOOT TH THEAD TITLE TR TT U UL VAR WBR
            ],
            procedure( x ); [^x ^x] endprocedure
        ).dup.length,
        false,
        "perm"
    )
enddefine;

define may_contain( parent, child );
    not( parent.is_optionally_closed ) or
    not( child.is_html ) or
    ( lmember( child, optional( parent ) ) and true )
enddefine;



;;; -- XML/HTML Parser ----------------------------------------


defclass lconstant Tag {
    tagType,
    tagAttributes
};

;;; This routine is used to mark where attribute values are
;;; constructed.  Really this is just a legacy component of this
;;; work.
;;;
define lconstant new_attribute =
    identfn
enddefine;

define lconstant read_white( ch, chars ) -> ch;
    while ch == `\s` or ch == `\t` or ch == `\n` or ch == `\r` do
        chars() -> ch
    endwhile;
    if ch == termin then
        mishap( 0, 'Unexpected end of input after "<"', 'eof-tag:xml-syntax' )
    endif;
enddefine;

define lconstant isnametokencode( ch );
    ch.isalphacode or ch.isnumbercode or locchar( ch, 1, '-_:.' )
enddefine;

define lconstant read_name( ch, chars ) -> ch;
    if isalphacode( ch ) then
        consword(#|
            ch;
            repeat
                chars() -> ch;
            quitunless( ch.isnametokencode );
                ch
            endrepeat
        |#)
    else
        mishap( consstring( ch, 1 ), 1, 'Incorrect character starting name', 'name:xml-syntax' )
    endif
enddefine;

define lconstant read_name_white( ch, chars );
    read_white( read_name( ch, chars ), chars )
enddefine;

;;; WRONG!  Attribute values are also subject to entity interpretation.
define lconstant read_string( ch, chars ) -> ch;
    consstring(#|
        if ch == `"` then
            chars() -> ch;
            until ch == `"` do
                if ch == termin do
                    mishap( 0, 'Unexpected end of input while reading attribute', 'eof-attribute:xml-syntax' )
                endif;
                ch; chars() -> ch;
            enduntil;
            chars() -> ch;
        elseif ch.isnametokencode do
            repeat
                ch;
            quitunless( isnametokencode( chars() ->> ch ) );
            endrepeat;
        else
            mishap( 0, 'Missing double-quote at start of attribute value', 'attribute:xml-syntax' )
        endif
    |#)
enddefine;

define lconstant read_string_white( ch, chars );
    read_white( read_string( ch, chars ), chars )
enddefine;

define lconstant ampersand( chars );

    define fetch( string, ans, chars );
        lvars x;
        for x in_string string do
            lvars ch = chars();
            unless x == ch do
                mishap( consstring( ch, 1 ), 1, 'Invalid character following "&"', 'entity:xml-syntax' )
            endunless
        endfor;
        ans.explode
    enddefine;

    chars() -> lvar ch;
    if ch == `a` then
        chars() -> ch;
        if ch == `m` then
            fetch( 'p;', '&', chars )
        elseif ch == `p` then
            fetch( 'os;', '\'', chars )
        else
            mishap( consstring( ch, 1 ), 1, 'Invalid character following "&a"', 'entity:xml-syntax' )
        endif
    elseif ch == `l` then
        fetch( 't;', '<', chars )
    elseif ch == `g` then
        fetch( 't;', '>', chars )
    elseif ch == `q` then
        fetch( 'uot;', '"', chars )
    elseif ch == `n` then
        fetch( 'bsp;', '&nbsp;', chars )
    else
        mishap( consstring( ch, 1 ), 1, 'Invalid character following "&"', 'entity:xml-syntax' )
    endif
enddefine;

define lconstant skip_comment( procedure chars );
    lvars n = 0;
    repeat
        lvars ch = chars();
        if ch == termin do
            mishap( 0, 'Unexpected end of input while processing comment', 'eof-comment:xml-syntax' )
        endif;
        if ch == `-` then
            n + 1 -> n;
            if n >= 2 then
                quitloop
            endif
        else
            0 -> n
        endif;
    endrepeat
enddefine;

define read_tag( ch, chars ) -> ch;
    lvars type, attributes;
    read_name_white( ch, chars ) -> ch -> type;                 ;;; returns type
    [%
        until ch == `/` or ch == `>` do
            new_attribute(
                read_name_white( ch, chars ) -> ch,         ;;; returns name
                if ch == `=` then
                    read_white( chars(), chars ) -> ch;
                    read_string_white( ch, chars ) -> ch;   ;;; returns value
                else
                    nullstring                              ;;; default
                endif
            )
        enduntil
    %] -> attributes;
    if ch == `/` then
        read_white( chars(), chars ) -> ch;
        if ch /== `>` then
            mishap( 0, 'Missing ">" following "/"', 'tag:xml-syntax' )
        endif;
        new_element( 0, type, attributes )
    elseif ch == `>` then
        consTag( type, attributes )
    else
        mishap( consstring( ch, 1 ), 1, 'Unclosed tag', 'tag:xml-syntax' )
    endif;
    chars() -> ch;
enddefine;

;;; read_token returns
;;;     a string, representing PCDATA
;;;     an Element, representing standalone/leaf tags
;;;     a Tag, representing an open tag
;;;     a word, representing an end-tag
;;;     termin, representing end of input
;;;
define read_token( ch, procedure chars ) -> ch;
    if ch == termin then
        termin
    elseif ch == `<` then
        read_white( chars(), chars ) -> ch;
        if ch == `/` then
            lblock
                read_name_white( chars(), chars ) -> ch -> lvar type;
                unless ch == `>` do
                    mishap( consstring( ch, 1 ), 1, 'End tag improperly closed', 'tag:xml-syntax' )
                endunless;
                chars() -> ch;
                type;
            endlblock
        elseif ch == `!` then
            chars() -> ch;
            if ch == `-` then
                unless chars() == `-` do
                    mishap( 0, 'Malformed comment (only one hypen)', 'comment:xml-syntax' )
                endunless;
                skip_comment( chars );
                read_white( chars(), chars ) -> ch;
                unless ch == `>` do
                    mishap( 0, 'Incorrectly terminated comment', 'comment:xml-syntax' )
                endunless;
                read_token( chars(), chars ) -> ch;
            else
                ;;; Just a temporary measure (I hope).
                until ( chars() ->> ch ) == termin or ch == `>` do enduntil;
                read_token( read_white( chars(), chars ), chars ) -> ch;
            endif
        else
            read_tag( ch, chars ) -> ch;
        endif
    elseif ch == `>` then
        mishap( 0, 'Misplaced ">"', 'tag:xml-syntax' )
    else
        consstring(#|
            repeat
                if ch == `&` then
                    ampersand( chars );
                else
                    ch
                endif;
                chars() -> ch;
                quitif( ch == termin );
                if ch == `<` or ch == `>` then
                    quitloop;
                endif;
            endrepeat
        |#)
    endif
enddefine;

define read_html_object( token_stream ) -> token_stream;
    lvars it = token_stream.dest -> token_stream;
    if it.isTag then
        it.tagType -> lvar type;
        if type.is_stand_alone then
            ;;; Eat optional end-tag.
            if not( null( token_stream ) ) and hd( token_stream ) == type then
                token_stream.tl -> token_stream
            endif;
            new_element( 0, it.tagType, it.tagAttributes )
        else
            new_element(
                #|
                    repeat
                        if null( token_stream ) then
                            mishap( type, 1, 'Unexpected end-of-input before end-tag', 'eof-tag:xml-syntax' )
                        endif;
                        lvars old_stream = token_stream;
                        lvars next = read_html_object( token_stream ) -> token_stream;
                        if next.is_element then
                            ;;; This may force closure.
                            if may_contain( type, next.element_type ) then
                                next
                            else
                                conspair( next, token_stream ) -> token_stream;
                                quitloop
                            endif
                        elseif next.isword then
                            ;;; This will force closure.
                            quitif( next == type );
                            if type.is_optionally_closed then
                                old_stream -> token_stream;
                                quitloop
                            else
                                mishap( next, 1, 'Misplaced end-tag', 'tag-end:xml-syntax' )
                            endif
                        else
                            next
                        endif;
                    endrepeat
                |#,
                type, it.tagAttributes
            )
        endif
    else
        it
    endif;
enddefine;

define read_xml_object( token_stream ) -> token_stream;
    lvars it = token_stream.dest -> token_stream;
    if it.isTag then
        new_element(
            #|
                repeat
                    lvars kid = read_xml_object( token_stream ) -> token_stream;
                    if kid.isword then
                        if kid == it.tagType then
                            quitloop
                        else
                            mishap( kid, 1, 'Mismatched end-tag', 'tag-end:xml-syntax' )
                        endif
                    endif;
                    kid
                endrepeat;
            |#,
            it.tagType, it.tagAttributes
        )
    else
        it
    endif
enddefine;

define lconstant open_source( s );
    if s.isstring or s.isdevice then
        discin( s )
    elseif s.isprocedure then
        s
    else
        mishap( s, 1, 'Invalid character source' )
    endif
enddefine;

define in_general_item( procedure char_source, read_obj );
    open_source( char_source ) -> char_source;

    lvars linenum = 1;
    define chars() -> ch;
        fast_apply( char_source ) -> ch;
        if ch == `\n` then
            1 + linenum -> linenum
        endif;
    enddefine;

    lvars ch = false;
    lvars token_stream =
        procedure();
            read_token( ch or chars(), chars ) -> ch
        endprocedure.pdtolist;

    procedure();

        define dlocal pop_exception_handler( N, mess, id, severity );
            if severity == `E` and hasendstring( id, ':xml-syntax' ) then
                sys_pr_message( N, sprintf( '%p (LINE %p)', [^mess ^linenum] ), id, severity );
                interrupt();
                setpop();
            else
                false
            endif
        enddefine;

        if null( token_stream ) then
            termin
        else
            lvars it = read_obj( token_stream ) -> token_stream;
            if it.isword then
                mishap( it, 1, 'Unmatched end-tag', 'tag-end:xml-syntax' )
            endif;
            it
        endif
    endprocedure
enddefine;

define in_xml_item =
    in_general_item(% read_xml_object %)
enddefine;

define in_html_item =
    in_general_item(% read_html_object %)
enddefine;

define read_xml_item =
    in_xml_item <> apply
enddefine;

define read_html_item =
    in_html_item <> apply
enddefine;

define read_xml_item_list =
    in_xml_item <> pdtolist
enddefine;

define read_html_item_list =
    in_html_item <> pdtolist
enddefine;



;;; -- Rendering XML/HTML items -------------------------------

constant procedure ( render_xml_item, render_html_item );

;;; WRONG! Must protect quotes!!
define lconstant print =
    pr
enddefine;

define lconstant renderAnyTag( t, e, start );
    appdata( start, cucharout );
    appdata( t, cucharout );
    if e then
        lvars name, value;
        for name, value in_attributes e do
            cucharout( ` ` );
            appdata( name, cucharout );
            cucharout( `=` );
            cucharout( `"` );
            print( value );
            cucharout( `"` );
        endfor
    endif;
    cucharout( `>` );
enddefine;

define lconstant renderTag( t, e );
    renderAnyTag( t, e, '<' )
enddefine;

define lconstant renderEndTag( t );
    renderAnyTag( t, false, '</' )
enddefine;

define lconstant renderXmlElement( e );
    lvars t = e.element_type;
    if t.is_stand_alone then
        renderTag( t, e )
    else
        renderTag( t, e );
        app_element_children( e, render_xml_item );
        renderEndTag( t );
    endif
enddefine;

define lconstant renderHtmlElement( e );
    lvars t = e.element_type;
    renderTag( t, e );
    app_element_children( e, render_html_item );
    renderEndTag( t );
enddefine;

define render_xml_item( x );
    if x.is_element then
        renderXmlElement( x )
    else
        print( x )
    endif
enddefine;

define render_html_item( x );
    if x.is_element then
        renderHtmlElement( x )
    else
        print( x )
    endif
enddefine;

endsection;
